import pandas as pd
import numpy as np
from scipy.stats import norm, t
import scipy.stats as st
from statsmodels.tsa.arima.model import ARIMA
from . import portfolio
from . import covarience_matrix
from . import simulation


# Using a fitted AR(1) model
def ar1_var_es(returns, size=10000, alpha=0.05):
    # Fit AR(1) model to returns
    ar1_model = ARIMA(returns, order=(1, 0, 0)).fit()
    
    # Extract model parameters and residuals
    ar1_coeff = ar1_model.params[0]
    residuals = ar1_model.resid
    sigma = np.std(residuals)
    simulated_returns = np.empty(size)
    returns = returns.values
    last_return = returns.iloc[-1]
    # Compute simulated returns using AR(1) model
    for i in range(size):
        simulated_returns[i] = ar1_coeff * last_return + sigma * np.random.normal()
    
    # Compute VaR and ES from simulated returns
    var_ar1 = -np.percentile(simulated_returns, alpha * 100)
    es_ar1 = -np.mean(simulated_returns[simulated_returns <= -var_ar1])
    
    return var_ar1, es_ar1, simulated_returns


# calculate var and es by using historical
def calculate_var_hist(returns, alpha=0.05):
    
    # Compute VaR using historic simulation
    var_hist = -np.percentile(returns, alpha * 100)
    
    # Compute ES using historic simulation and VaR
    es_hist = -np.mean(returns[returns < -var_hist])
    
    # Return VaR, ES, and original returns
    return var_hist, es_hist, returns



# calculate normal var and es
def noramal_VaR_ES(data, size=10000, alpha=0.05):

    # Calculate mean and standard deviation of historical returns
    mu = data.mean().values
    sigma = np.std(data).values

    # Generate simulated returns using the normal distribution with mean and standard deviation
    simu_return = np.random.normal(mu, sigma, size)

    # Sort the simulated returns in ascending order
    simu_return.sort()

    # Calculate the index of the value at risk at the confidence level alpha
    n = int(alpha * len(simu_return))

    # Calculate the indices of the upper and lower values at risk
    iup = int(np.ceil(n))
    idn = int(np.floor(n))

    # Calculate the Value at Risk (VaR) as the average of the upper and lower values
    VaR = -(simu_return[iup] + simu_return[idn])/2

    # Calculate the Expected Shortfall (ES) as the average of all returns below the lower value at risk
    ES = -np.mean(simu_return[0:idn])

    # Return VaR, ES, and the sorted simulated returns
    return VaR, ES, simu_return



# Using a normal distribution with an Exponentially Weighted variance ( )λ=0.94
def normal_weighted(meta ,size=10000,lambda_=0.94, alpha=0.05):
    mean_data=meta.mean()
    sigma = np.sqrt(covarience_matrix.exp_weighted_cov(meta,lambda_=lambda_))
    sigma=sigma[0][0]
    simu_norm = np.random.normal(mean_data,sigma,size)
    var_ew = -np.percentile(simu_norm,alpha*100)
    es_ew = -np.mean(simu_norm[simu_norm<=var_ew])

    return  var_ew,es_ew,simu_norm

# MLE fitted T distribution
def t_VaR_ES(data, alpha = 0.05,size=10000):

    t_df, t_m, t_s = st.t.fit(data)
    t = st.t.rvs(df = t_df, loc = t_m, scale = t_s, size=size)
    tsim = pd.DataFrame({"tsim": t})
    VaR_T = -st.t.ppf(alpha, df = t_df, loc = t_m, scale = t_s)
    temp = tsim[tsim <= -VaR_T].dropna()
    ES_T = -temp.mean().values
    return VaR_T, ES_T[0] ,t


# Delta Normal VaR
def Delta_Normal_VaR(portfolio_type,prices,lambda_=0.94,portfolio_='All'):
    current_price, assets_prices, hold = portfolio.Norm_price(portfolio_type,prices,portfolio_=portfolio_)
    ret = portfolio.return_calculate(assets_prices).drop('Date', axis=1)
    cov = covarience_matrix.exp_weighted_cov(ret,lambda_)
    sigm = np.sqrt(np.transpose(hold) @ cov @ hold)
    var = -current_price * norm.ppf(0.05) * sigm


    return current_price[0], var[0][0]



# Normal Monte Carlo vaR

def cal_monte_var(portfolio_type,prices, n=10000,lambda_=0.94,alpha=0.05,portfolio_='All'):
    
    current_price, assets_prices, hold = portfolio.cal_porfolio_price(portfolio_type,prices,portfolio_=portfolio_)


    re_n = portfolio.return_calculate(assets_prices).drop('Date', axis=1)

    re_no = re_n - re_n.mean()

    cov = covarience_matrix.exp_weighted_cov(re_no,lambda_)
  
    np.random.seed(0)
    return_simu = np.add(simulation.multivariate_normal_simulation(cov,n),re_n.mean().values)
    assets_prices = assets_prices.drop('Date',axis=1)
    sp=np.dot(return_simu*assets_prices.tail(1).values.reshape(assets_prices.shape[1],),hold)
    var = -np.percentile(sp, alpha*100)
    es = -np.mean(sp[sp<=-var])

    return current_price[0], var,es,sp

# historical var
def cal_historical(portfolio_type, prices,n=10000,alpha=0.05,portfolio_='All'):
    
    current_price, assets_prices, hold = portfolio.cal_porfolio_price(portfolio_type,prices,portfolio_=portfolio_)
    ret = portfolio.return_calculate(assets_prices).drop('Date', axis=1)
    
    assets_prices = assets_prices.drop('Date',axis=1)
    simu = ret.sample(n,replace=True)
    his_pri = np.dot(simu * assets_prices.tail(1).values.reshape(assets_prices.shape[1]), hold)
    

    var_historical = -np.percentile(his_pri, alpha*100)
    es_historical = -np.mean(his_pri[his_pri <=-var_historical])

    return current_price[0], var_historical, es_historical,his_pri


import numpy as np
import pandas as pd


# calculate weight covariance
def preprocess_returns(returns):

    returns_array = returns.values
    mean_return = np.mean(returns_array, axis=0)
    normalized_returns = returns_array - mean_return
    return normalized_returns, mean_return


def exp_weighted_cov(returns, lambda_=0.97):
    
    normalized_returns, mean_return = preprocess_returns(returns)
    n = normalized_returns.shape[0]
    weighted_cov = np.cov(normalized_returns.T)
    for t in range(1, n):
        current_return = normalized_returns[t]
        weighted_cov = lambda_ * weighted_cov + (1 - lambda_) * np.outer(current_return, current_return)
    return weighted_cov

# calculate weighted matrxi

def exp_weighted_matrix(returns, lambda_=0.97):

    returns = returns.values
    n_w_m = returns.shape[0]
    weights = np.zeros(n_w_m)
    
    for t in range(n_w_m):
        weights[n_w_m-1-t]  = (1-lambda_)*lambda_**t

    weights_matrix = np.diag(weights/sum(weights))

    return weights_matrix

# cholesky decomposition
def chol_psd(matrix):
    le = len(matrix)
    ze = -1e-8
    root = np.array([[0.0] * le for _ in range(le)])
    for j in range(le):
        s = root[j,:j]@root[j,:j].T
        temp = matrix[j,j] - s
        if 0 >= temp >= ze:
            temp = 0.0
        elif temp < ze:
            raise ValueError("non-PSD")
        root[j,j] = np.sqrt(temp)
        if root[j,j] == 0:
            continue
        for i in range(j+1,le):
            s = root[i,:j]@root[j,:j]
            root[i,j] =(matrix[i,j] - s)*1.0/root[j,j]
    return root


# Rebonato and Jackel
def near_psd(matrix, epsilon=0.0):

    invSD = None

    n = matrix.shape[0]
    out = matrix.copy()
    if np.count_nonzero(np.diag(out) == 1.0) != n:
        invSD = np.diag(1 / np.sqrt(np.diag(out)))
        out = np.matmul(np.matmul(invSD, out), invSD)

    vals, vecs = np.linalg.eigh(out)
    vals = np.maximum(vals, epsilon)
    T = np.reciprocal(np.matmul(np.square(vecs), vals))
    T = np.diag(np.sqrt(T))
    l = np.diag(np.sqrt(vals))
    B = np.matmul(np.matmul(T, vecs), l)
    out = np.matmul(B, np.transpose(B))

    if invSD is not None:
        invSD = np.diag(1 / np.diag(invSD))
        out = np.matmul(np.matmul(invSD, out), invSD)

    return out

# Higham
def proj_u(matrix):
    cor = matrix.copy()
    np.fill_diagonal(cor,1)
    return cor

def Frobenius_norm(matrix):
    return np.linalg.norm(matrix, 'fro')**2

def proj_s(matrix):
    eig_val, eig_vec = np.linalg.eigh(matrix)
    eig_val[eig_val<0] = 0
    p = eig_vec@ np.diagflat(eig_val)@eig_vec.T
    return p


def Higham_near_psd(matrix):
    dS = 0
    Y = matrix
    last_gamma = float("inf")
    iteration = 100000
    tol = 1e-10

    for i in range(iteration):
        R = Y - dS                      
        X = proj_s(R)                   
        dS = X - R                       
        Y = proj_u(X)                   
        gamma = Frobenius_norm(Y - matrix)   
        if abs(gamma-last_gamma)< tol:  
            break
        last_gamma = gamma
    return Y


import numpy as np
from . import covarience_matrix

def multivariate_normal_simulation(covariance_matri, n_samples, explained_variance=1.0,mean=0, method="Direct"):
    if method == "PCA":

        eigenvalues, eigenvectors = np.linalg.eigh(covariance_matri)

        mask = eigenvalues > np.finfo(np.float32).eps
        eigenvalues = eigenvalues[mask][::-1]
        eigenvectors = eigenvectors[:, mask][:, ::-1]

        if explained_variance == 1.0:
            explained_variance = 1.0
        elif explained_variance <= 0.0:
            raise ValueError("explained_variance must be a positive number")
        else:
            explained_variance = min(1.0, explained_variance)

        cum_var = np.cumsum(eigenvalues) / np.sum(eigenvalues)
        n_components = np.searchsorted(
        cum_var, explained_variance, side='right')
        z = np.random.normal(size=(n_components, n_samples))
        B = eigenvectors[:, :n_components] * \
          np.sqrt(eigenvectors[:n_components])
        x = np.dot(B, z).T

        return x
    elif method == "Direct":
        ma = covarience_matrix.chol_psd(covariance_matri)
        n=np.random.normal(size=(covariance_matri.shape[0], n_samples))
        x=np.transpose(np.dot(ma, n) + mean)
        return x






import pandas as pd
import numpy as np

# portfolio A/B/C
def cal_porfolio_price(portfolio,prices,portfolio_="All"):
    if portfolio_=="All":
        a=portfolio.drop('Portfolio',axis=1)
        portfolio_type = a.groupby('Stock', as_index=False).sum('Holding')
    else:
        portfolio_type = portfolio[portfolio["Portfolio"]==portfolio_]
    stock_name = list(portfolio_type["Stock"])
    assets_prices = pd.concat([prices["Date"], prices[stock_name]], axis=1)
    current_price = np.dot(prices[portfolio_type["Stock"]].tail(1), portfolio_type["Holding"])
    holdings = portfolio_type["Holding"]

    return current_price,assets_prices,holdings

# all 
def Norm_price(portfolio,prices,portfolio_='All'):
    if portfolio_=="All":
        a=portfolio.drop('Portfolio',axis=1)
        portfolio_type = a.groupby('Stock', as_index=False).sum('Holding')
    else:
        portfolio_type = portfolio[portfolio["Portfolio"]==portfolio_]
        
    stock_name = list(portfolio_type["Stock"])
    assets_prices = pd.concat([prices["Date"], prices[stock_name]], axis=1)
    current_price = np.dot(prices[portfolio_type["Stock"]].tail(1), portfolio_type["Holding"])

    val = portfolio_type["Holding"].values.reshape(-1,1) * prices[portfolio_type["Stock"]].tail(1).T.values
    hold = val/current_price
   

    return current_price,assets_prices,hold


# specify the method of return calculation
def return_calculate(prices, method = 'Arithmetic'):
    prices_date = prices.Date[1:]
    row = len(prices_date)

    col = prices.columns
    col_new = col.drop('Date')
    le = len(col_new)

    price_new = prices[col_new].to_numpy()

    price_return = np.empty((row,le))

    for i in range(row):
        for j in range(le):
            price_return[i][j] = price_new[i+1][j]/price_new[i][j]


    if method == 'Arithmetic':
        price_return = price_return - 1.0
    elif method == 'Log':
        price_return = np.log(price_return)
    else:
        raise ValueError(f'Wrong method')

    dateColumn = "Date"
    prices_da = pd.DataFrame({dateColumn: prices_date}) 

    for i in range(le):
        prices_da[col_new[i]] = price_return[:,i]
    
    return prices_da


from setuptools import setup, find_packages
import os
setup(
    name="mjzy",
    version="0.3.7",
    author="Meijing Hou",
    author_email="houmeijing15@163.com",
    description="risk management",
    long_description="fintech545 assignment4",
    long_description_content_type="text/markdown",
    url="https://github.com/meijinghou/meijing-hou-545",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',

)
